import asyncio
import logging
import uuid
from typing import AsyncIterator, Optional

from ai.backend.common.events.fetcher import EventFetcher
from ai.backend.logging.utils import BraceStyleAdapter

from ...dispatcher import AbstractEvent
from ..hub import EventPropagator

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class WithCachePropagator(EventPropagator):
    """
    WithCachePropagator is an event propagator that handles broadcast events.
    It fetches the last event for a given task ID from the cache and propagates events
    to a queue for broadcasted events.
    It allows to prevent timing issues by ensuring that the last event is fetched before yielding new events.
    """

    _id: uuid.UUID
    _event_fetcher: EventFetcher
    _queue: asyncio.Queue[Optional[AbstractEvent]]
    _closed: bool = False

    def __init__(self, event_fetcher: EventFetcher) -> None:
        self._id = uuid.uuid4()
        self._event_fetcher = event_fetcher
        self._queue = asyncio.Queue()
        self._closed = False

    def id(self) -> uuid.UUID:
        """
        Get the unique identifier for the propagator.
        """
        return self._id

    async def receive(self, cache_id: str) -> AsyncIterator[AbstractEvent]:
        """
        First, it fetches the last event for a given cache ID using the event fetcher.
        If the last event is not None, it yields that event.
        Then, it enters a loop to receive events from the queue until closed.
        """
        cached_event = await self._event_fetcher.fetch_cached_event(cache_id)
        if cached_event is not None:
            yield cached_event
        while not self._closed:
            event = await self._queue.get()
            try:
                if event is None:
                    break
                yield event
            except Exception as e:
                log.error("Error propagating event: {}", e)

    async def propagate_event(self, event: AbstractEvent) -> None:
        await self._queue.put(event)

    async def close(self) -> None:
        if self._closed:
            return
        self._closed = True
        await self._queue.put(None)



import json
import uuid
from dataclasses import dataclass
from typing import Any, Sequence

from redis.asyncio.client import Pipeline, Redis

from ai.backend.common import redis_helper
from ai.backend.common.json import dump_json_str
from ai.backend.common.types import RedisConnectionInfo

from ..service_discovery import ServiceDiscovery, ServiceMetadata

_DEFAULT_PREFIX = "service_discovery"
_DEFAULT_TTL = 60 * 3  # 3 minutes


@dataclass
class RedisServiceDiscoveryArgs:
    redis: RedisConnectionInfo
    ttl: int = _DEFAULT_TTL  # 3 minutes
    prefix: str = _DEFAULT_PREFIX


class RedisServiceDiscovery(ServiceDiscovery):
    _redis: RedisConnectionInfo
    _ttl: int
    _prefix: str

    def __init__(self, args: RedisServiceDiscoveryArgs) -> None:
        self._redis = args.redis
        self._ttl = args.ttl
        self._prefix = args.prefix

    async def _hsetex(self, key: str, mapping: dict[str, Any]) -> None:
        # TODO: Use actual `hsetex` after upgrading to redis-py 8.0
        async def _pipe_builder(r: Redis) -> Pipeline:
            pipe = r.pipeline()
            await pipe.hset(key, mapping={k: dump_json_str(v) for k, v in mapping.items()})
            await pipe.expire(key, self._ttl)
            await pipe.execute()
            return pipe

        await redis_helper.execute(self._redis, _pipe_builder)

    async def _scan_keys(self, pattern: str) -> list[str]:
        async def _run(r: Redis) -> list[str]:
            return [key.decode() async for key in r.scan_iter(match=pattern, count=100)]

        return await redis_helper.execute(self._redis, _run)

    async def _hget_json(self, name: str) -> ServiceMetadata:
        raw_hash: dict[bytes, bytes] = await redis_helper.execute(
            self._redis, lambda r: r.hgetall(name)
        )
        if not raw_hash:
            raise ValueError(f"Service key {name} not found.")

        result: dict[str, Any] = {}
        for raw_key, raw_value in raw_hash.items():
            key = raw_key.decode()
            value = json.loads(raw_value)
            result[key] = value

        return ServiceMetadata.from_dict(result)

    async def register(self, service_meta: ServiceMetadata) -> None:
        key = self._service_prefix(service_meta.service_group, service_meta.id)
        await self._hsetex(key, service_meta.to_dict())

    async def unregister(self, service_group: str, service_id: uuid.UUID) -> None:
        key = self._service_prefix(service_group, service_id)
        await redis_helper.execute(
            self._redis,
            lambda r: r.delete(key),
            service_name=self._redis.service_name,
        )

    async def heartbeat(self, service_meta: ServiceMetadata) -> None:
        service_meta.health_status.update_heartbeat()
        key = self._service_prefix(service_meta.service_group, service_meta.id)
        await self._hsetex(key, service_meta.to_dict())

    async def discover(self) -> Sequence[ServiceMetadata]:
        pattern = f"{self._prefix}.*.*"
        keys = await self._scan_keys(pattern)
        if not keys:
            raise ValueError("No service groups found.")
        return [await self._hget_json(key) for key in keys]

    async def get_service_group(self, service_group: str) -> Sequence[ServiceMetadata]:
        pattern = f"{self._service_group_prefix(service_group)}.*"
        keys = await self._scan_keys(pattern)
        if not keys:
            raise ValueError(f"No services found in group {service_group}.")
        return [await self._hget_json(key) for key in keys]

    async def get_service(self, service_group: str, service_id: uuid.UUID) -> ServiceMetadata:
        key = self._service_prefix(service_group, service_id)
        return await self._hget_json(key)

    def _service_group_prefix(self, service_group: str) -> str:
        return f"{self._prefix}.{service_group}"

    def _service_prefix(self, service_group: str, service_id: uuid.UUID) -> str:
        return f"{self._service_group_prefix(service_group)}.{service_id}"



import base64
from dataclasses import dataclass
from typing import Mapping, Self, cast

from ai.backend.common import msgpack

_DEFAULT_RETRY_FIELD = b"_retry_count"
_DEFAULT_MAX_RETRIES = 3

type MessageId = bytes


@dataclass
class BroadcastMessage:
    payload: Mapping[str, str]


@dataclass
class MQMessage:
    msg_id: MessageId
    payload: dict[bytes, bytes]

    def retry(self) -> bool:
        """
        Retry the message.
        If the message has been retried more than the maximum number of retries,
        the message will be discarded.
        The retry count is stored in the message payload.
        """
        if self._retry_count() > _DEFAULT_MAX_RETRIES:
            return False
        self.payload[_DEFAULT_RETRY_FIELD] = str(self._retry_count() + 1).encode("utf-8")
        return True

    def _retry_count(self) -> int:
        """
        Get the retry count of the message.
        The retry count is the number of times the message has been re-delivered.
        """
        return int(self.payload.get(_DEFAULT_RETRY_FIELD, b"0"))


@dataclass
class MessagePayload:
    name: str
    source: str
    args: tuple[bytes, ...]

    def serialize_anycast(self) -> Mapping[bytes, bytes]:
        """
        Serialize the message payload to a dictionary.
        """
        return {
            b"name": self.name.encode(),
            b"source": self.source.encode(),
            b"args": msgpack.packb(self.args),
        }

    def serialize_broadcast(self) -> Mapping[str, str]:
        """
        Serialize the message payload to a dictionary.
        The keys are bytes and the values are bytes.
        """
        args = base64.b64encode(msgpack.packb(self.args)).decode("ascii")
        return {
            "name": self.name,
            "source": self.source,
            "args": args,
        }

    @classmethod
    def from_anycast(cls, payload: Mapping[bytes, bytes]) -> Self:
        """
        Deserialize the message payload from a dictionary.
        The keys are bytes and the values are bytes.
        """
        return cls(
            name=payload[b"name"].decode("utf-8"),
            source=payload[b"source"].decode("utf-8"),
            args=cast(tuple[bytes, ...], msgpack.unpackb(payload[b"args"])),
        )

    @classmethod
    def from_broadcast(cls, payload: Mapping[str, str]) -> Self:
        """
        Deserialize the message payload from a dictionary.
        The keys are bytes and the values are bytes.
        """
        args_bytes = base64.b64decode(payload["args"])
        return cls(
            name=payload["name"],
            source=payload["source"],
            args=cast(tuple[bytes, ...], msgpack.unpackb(args_bytes)),
        )



import asyncio
import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Iterable, Optional, Self

import glide
from glide import (
    GlideClient,
    GlideClientConfiguration,
    Logger,
    LogLevel,
    NodeAddress,
    ServerCredentials,
)
from redis.asyncio.sentinel import Sentinel

from ai.backend.logging import BraceStyleAdapter

from ...types import HostPortPair, RedisTarget
from ...validators import DelimiterSeperatedList
from ...validators import HostPortPair as _HostPortPair

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


_DEFAULT_REQUEST_TIMEOUT = 1_000  # Default request timeout in milliseconds

Logger.init(LogLevel.OFF)  # Disable Glide logging by default


@dataclass
class ValkeyStandaloneTarget:
    address: HostPortPair
    password: Optional[str] = None
    request_timeout: Optional[int] = None

    @classmethod
    def from_redis_target(cls, redis_target: RedisTarget) -> Self:
        """
        Create a ValkeyConnectionConfig from a RedisTarget.
        """
        if not redis_target.addr:
            raise ValueError("RedisTarget must have an address for standalone mode")
        return cls(
            address=redis_target.addr,
            password=redis_target.password,
            request_timeout=1_000,  # Default request timeout
        )


def _sentinel_from_redis_target(
    redis_target: RedisTarget,
) -> Optional[Iterable[HostPortPair]]:
    if isinstance(redis_target.sentinel, str):
        return DelimiterSeperatedList(_HostPortPair).check_and_return(redis_target.sentinel)
    elif isinstance(redis_target.sentinel, list):
        return redis_target.sentinel
    return None


@dataclass
class ValkeySentinelTarget:
    sentinel_addresses: Iterable[HostPortPair]
    service_name: str
    password: Optional[str] = None
    request_timeout: Optional[int] = None

    @classmethod
    def from_redis_target(cls, redis_target: RedisTarget) -> Self:
        """
        Create a ValkeySentinelTarget from a RedisTarget.
        """
        if not redis_target.service_name:
            raise ValueError("RedisTarget must have service_name when using sentinel")
        # Convert sentinel addresses to list format
        sentinel_addresses = _sentinel_from_redis_target(redis_target)
        if not sentinel_addresses:
            raise ValueError("RedisTarget sentinel configuration is invalid or empty")

        return cls(
            sentinel_addresses=sentinel_addresses,
            service_name=redis_target.service_name,
            password=redis_target.password,
            request_timeout=1_000,  # Default request timeout
        )


class AbstractValkeyClient(ABC):
    @property
    @abstractmethod
    def client(self) -> GlideClient:
        pass

    @abstractmethod
    async def connect(self) -> None:
        pass

    @abstractmethod
    async def disconnect(self) -> None:
        pass


class ValkeyStandaloneClient(AbstractValkeyClient):
    _target: ValkeyStandaloneTarget
    _valkey_client: Optional[GlideClient]
    _db_id: int
    _pubsub_channels: Optional[set[str]]
    _human_readable_name: str
    _monitor_task: Optional[asyncio.Task[None]]

    def __init__(
        self,
        target: ValkeyStandaloneTarget,
        db_id: int,
        human_readable_name: str,
        pubsub_channels: Optional[set[str]] = None,
    ) -> None:
        self._target = target
        self._valkey_client = None
        self._db_id = db_id
        self._human_readable_name = human_readable_name
        self._pubsub_channels = pubsub_channels
        self._monitor_task = None

    @property
    def client(self) -> GlideClient:
        if self._valkey_client is None:
            raise RuntimeError("ValkeyStandaloneClient not connected. Call connect() first.")
        return self._valkey_client

    async def connect(self) -> None:
        if self._valkey_client is not None:
            return

        await self._create_valkey_client()
        self._monitor_task = asyncio.create_task(self._monitor_connection())

    async def disconnect(self) -> None:
        if self._monitor_task:
            self._monitor_task.cancel()
            try:
                await self._monitor_task
            except asyncio.CancelledError:
                pass
            self._monitor_task = None

        if self._valkey_client:
            await self._valkey_client.close(err_message="ValkeyStandaloneClient is closed.")
            self._valkey_client = None

    async def _create_valkey_client(self) -> None:
        target_host = self._target.address.host
        target_port = self._target.address.port
        addresses = [NodeAddress(host=target_host, port=target_port)]

        credentials = (
            ServerCredentials(password=self._target.password) if self._target.password else None
        )

        config = GlideClientConfiguration(
            addresses,
            credentials=credentials,
            database_id=self._db_id,
            request_timeout=self._target.request_timeout or _DEFAULT_REQUEST_TIMEOUT,
            pubsub_subscriptions=GlideClientConfiguration.PubSubSubscriptions(
                channels_and_patterns={
                    GlideClientConfiguration.PubSubChannelModes.Exact: self._pubsub_channels,
                },
                callback=None,
                context=None,
            )
            if self._pubsub_channels
            else None,
        )

        glide_client = await GlideClient.create(config)
        self._valkey_client = glide_client

        log.info(
            "Created ValkeyClient for standalone at {}:{} for database {}",
            target_host,
            target_port,
            self._human_readable_name,
        )

    async def _ping(self) -> bool:
        """
        Ping the server to check if the connection is alive.
        """
        if self._valkey_client is None:
            return False
        try:
            await self._valkey_client.ping()
            return True
        except glide.ClosingError as e:
            log.warning(
                "Valkey client is closed for standalone at {}:{}, human readable name '{}': {}",
                self._target.address.host,
                self._target.address.port,
                self._human_readable_name,
                e,
            )
            return False
        except Exception as e:
            log.debug(
                "Failed to ping to service '{}', human readable name '{}', but cannot check if the connection is alive: {}",
                self._target.address,
                self._human_readable_name,
                e,
            )
            return True

    async def _check_connection(self) -> bool:
        """
        Check if the current connection is alive.
        If not, return False to trigger reconnection.
        """
        if not await self._ping():
            return False
        return True

    async def _monitor_connection(self) -> None:
        while True:
            try:
                await asyncio.sleep(10.0)
                if await self._check_connection():
                    continue
                await self._reconnect()

            except asyncio.CancelledError:
                break
            except Exception as e:
                log.exception("Error in connection monitoring: {}", e)

    async def _reconnect(self) -> None:
        if self._valkey_client:
            try:
                await self._valkey_client.close()
            except Exception as e:
                log.warning("Error closing old client: {}", e)
            self._valkey_client = None

        await self._create_valkey_client()


class ValkeySentinelClient(AbstractValkeyClient):
    _target: ValkeySentinelTarget
    _db_id: int
    _human_readable_name: str
    _sentinel: Sentinel
    _pubsub_channels: Optional[set[str]]
    _valkey_client: Optional[GlideClient]
    _master_address: Optional[tuple[str, int]]
    _monitor_task: Optional[asyncio.Task[None]]

    def __init__(
        self,
        target: ValkeySentinelTarget,
        db_id: int,
        human_readable_name: str,
        pubsub_channels: Optional[set[str]] = None,
    ) -> None:
        sentinel_addrs = []
        for addr in target.sentinel_addresses:
            sentinel_addrs.append((addr.host, addr.port))

        self._sentinel = Sentinel(
            sentinel_addrs,
            sentinel_kwargs={
                "password": target.password,
            },
        )
        self._db_id = db_id
        self._human_readable_name = human_readable_name
        self._target = target
        self._pubsub_channels = pubsub_channels
        self._valkey_client = None
        self._master_address = None
        self._monitor_task = None

    @property
    def client(self) -> GlideClient:
        if self._valkey_client is None:
            raise RuntimeError("ValkeySentinelClient not connected. Call connect() first.")
        return self._valkey_client

    async def connect(self) -> None:
        if self._valkey_client is not None:
            return

        await self._create_valkey_client()
        self._monitor_task = asyncio.create_task(self._monitor_connection())

    async def disconnect(self) -> None:
        if self._monitor_task:
            self._monitor_task.cancel()
            try:
                await self._monitor_task
            except asyncio.CancelledError:
                pass
            self._monitor_task = None

        if self._valkey_client:
            await self._valkey_client.close(err_message="ValkeySentinelClient is closed.")
            self._valkey_client = None

    async def _create_valkey_client(self) -> None:
        master_address = await self._get_master_address()
        if master_address is None:
            raise RuntimeError(
                f"Cannot find master for service '{self._target.service_name}' in {self._human_readable_name}"
            )

        self._master_address = master_address

        addresses = [NodeAddress(host=master_address[0], port=master_address[1])]
        credentials = (
            ServerCredentials(password=self._target.password) if self._target.password else None
        )

        config = GlideClientConfiguration(
            addresses,
            credentials=credentials,
            database_id=self._db_id,
            client_name=self._target.service_name,
            request_timeout=self._target.request_timeout or _DEFAULT_REQUEST_TIMEOUT,
            pubsub_subscriptions=GlideClientConfiguration.PubSubSubscriptions(
                channels_and_patterns={
                    GlideClientConfiguration.PubSubChannelModes.Exact: self._pubsub_channels,
                },
                callback=None,
                context=None,
            )
            if self._pubsub_channels
            else None,
        )

        glide_client = await GlideClient.create(config)
        self._valkey_client = glide_client

        log.info(
            "Created ValkeyClient for master at {}:{} for database {}",
            master_address[0],
            master_address[1],
            self._human_readable_name,
        )

    async def _get_master_address(self) -> Optional[tuple[str, int]]:
        try:
            return await self._sentinel.discover_master(self._target.service_name)
        except Exception as e:
            log.error(
                "Failed to discover master for service '{}': {}", self._target.service_name, e
            )
            return None

    async def _ping(self) -> bool:
        """
        Ping the current master to check if the connection is alive.
        """
        if self._valkey_client is None:
            return False
        try:
            await self._valkey_client.ping()
            return True
        except glide.ClosingError as e:
            log.warning(
                "Valkey client is closed for service '{}', human readable name '{}': {}",
                self._target.service_name,
                self._human_readable_name,
                e,
            )
            return False
        except Exception as e:
            log.debug(
                "Failed to ping to service '{}', human readable name '{}', but cannot check if the connection is alive: {}",
                self._target.service_name,
                self._human_readable_name,
                e,
            )
            return True

    async def _check_connection(self) -> bool:
        """
        Check if the current master connection is alive.
        If not, attempt to reconnect.
        """
        if not await self._ping():
            log.warning(
                "Connection to master {}:{} is down, attempting to reconnect",
                self._master_address[0] if self._master_address else "unregistered",
                self._master_address[1] if self._master_address else "unregistered",
            )
            return False
        current_master = await self._get_master_address()
        if current_master is None or current_master == self._master_address:
            return True
        log.warning(
            "Master change detected for service '{}': {}:{} -> {}:{} in {}",
            self._target.service_name,
            self._master_address[0] if self._master_address else "unregistered",
            self._master_address[1] if self._master_address else "unregistered",
            current_master[0],
            current_master[1],
            self._human_readable_name,
        )
        return False

    async def _monitor_connection(self) -> None:
        while True:
            try:
                await asyncio.sleep(10.0)
                if await self._check_connection():
                    continue
                log.info("Reconnecting to new master for service '{}'", self._target.service_name)
                await self._reconnect_to_new_master()

            except asyncio.CancelledError:
                break
            except Exception as e:
                log.exception("Error in master monitoring: {}", e)

    async def _reconnect_to_new_master(self) -> None:
        if self._valkey_client:
            try:
                await self._valkey_client.close()
            except Exception as e:
                log.warning("Error closing old client: {}", e)
            self._valkey_client = None

        await self._create_valkey_client()


def create_valkey_client(
    target: RedisTarget,
    db_id: int,
    human_readable_name: str,
    pubsub_channels: Optional[set[str]] = None,
) -> AbstractValkeyClient:
    """
    Factory function to create a Valkey client based on the target type.
    """
    if target.sentinel:
        sentinel_target = ValkeySentinelTarget.from_redis_target(target)
        return ValkeySentinelClient(sentinel_target, db_id, human_readable_name, pubsub_channels)
    standalone_target = ValkeyStandaloneTarget.from_redis_target(target)
    return ValkeyStandaloneClient(standalone_target, db_id, human_readable_name, pubsub_channels)



"""
Sample configuration generator for Backend.AI.

This module provides utilities to generate sample TOML configuration files
from Pydantic models by extracting field information, defaults, descriptions,
and examples using JSON Schema.
"""

import textwrap
from typing import Any, Optional, Type

import toml
from pydantic import BaseModel


def _wrap_comment(text: str, width: int = 80) -> str:
    """Wrap text into multiline comment format."""
    lines = text.strip().split("\n")
    wrapped_lines = []

    for line in lines:
        line = line.strip()
        if not line:
            wrapped_lines.append("#")
        else:
            # Wrap long lines
            wrapped = textwrap.wrap(line, width=width - 2)  # Account for '# '
            for wrapped_line in wrapped:
                wrapped_lines.append(f"# {wrapped_line}")

    return "\n".join(wrapped_lines)


def _get_field_comment(field_schema: dict[str, Any]) -> str:
    """Generate comment for a field based on its schema."""
    comment_parts = []

    # Add description
    if "description" in field_schema:
        comment_parts.append(field_schema["description"].strip())

    # Add examples
    if "examples" in field_schema:
        examples_str = ", ".join(str(ex) for ex in field_schema["examples"])
        comment_parts.append(f"Examples: {examples_str}")

    return "\n".join(comment_parts)


def _get_default_from_schema(field_schema: dict[str, Any]) -> Any:
    """Get default value from JSON schema."""
    if "default" in field_schema:
        return field_schema["default"]

    # Handle different types with appropriate defaults
    field_type = field_schema.get("type")
    if field_type == "string":
        return ""
    elif field_type == "integer":
        return 0
    elif field_type == "number":
        return 0.0
    elif field_type == "boolean":
        return False
    elif field_type == "array":
        return []
    elif field_type == "object":
        return {}

    return None


def _generate_config_from_schema(
    schema: dict[str, Any], definitions: Optional[dict[str, Any]] = None
) -> dict[str, Any]:
    """Generate configuration dictionary from JSON schema."""
    if definitions is None:
        definitions = schema.get("$defs", {})

    config: dict[str, Any] = {}
    properties = schema.get("properties", {})

    for field_name, field_schema in properties.items():
        # Handle references to other schemas
        if "$ref" in field_schema:
            ref_name = field_schema["$ref"].split("/")[-1]
            if ref_name in definitions:
                ref_schema = definitions[ref_name]
                # Only recurse if it's an object type (likely a BaseModel)
                if ref_schema.get("type") == "object":
                    config[field_name] = _generate_config_from_schema(ref_schema, definitions)
                else:
                    config[field_name] = _get_default_from_schema(ref_schema)
            else:
                config[field_name] = None

        # Handle allOf (used for inheritance)
        elif "allOf" in field_schema:
            merged_schema = {}
            for sub_schema in field_schema["allOf"]:
                if "$ref" in sub_schema:
                    ref_name = sub_schema["$ref"].split("/")[-1]
                    if ref_name in definitions:
                        ref_schema = definitions[ref_name]
                        merged_schema.update(ref_schema)
                else:
                    merged_schema.update(sub_schema)

            if merged_schema.get("type") == "object":
                config[field_name] = _generate_config_from_schema(merged_schema, definitions)
            else:
                config[field_name] = _get_default_from_schema(merged_schema)

        # Handle anyOf/oneOf (used for Union types)
        elif "anyOf" in field_schema or "oneOf" in field_schema:
            variants = field_schema.get("anyOf", field_schema.get("oneOf", []))

            # Find the first non-null variant
            for variant in variants:
                if variant.get("type") != "null":
                    if "$ref" in variant:
                        ref_name = variant["$ref"].split("/")[-1]
                        if ref_name in definitions:
                            ref_schema = definitions[ref_name]
                            if ref_schema.get("type") == "object":
                                config[field_name] = _generate_config_from_schema(
                                    ref_schema, definitions
                                )
                            else:
                                config[field_name] = _get_default_from_schema(ref_schema)
                        else:
                            config[field_name] = None
                    else:
                        if variant.get("type") == "object":
                            config[field_name] = _generate_config_from_schema(variant, definitions)
                        else:
                            config[field_name] = _get_default_from_schema(variant)
                    break
            else:
                # All variants are null or no suitable variant found
                config[field_name] = None

        # Handle object type (nested models)
        elif field_schema.get("type") == "object":
            config[field_name] = _generate_config_from_schema(field_schema, definitions)

        # Handle primitive types
        else:
            config[field_name] = _get_default_from_schema(field_schema)

    return config


def _add_field_comments_from_schema(
    toml_str: str, schema: dict[str, Any], definitions: Optional[dict[str, Any]] = None
) -> str:
    """Add comments to TOML string based on JSON schema."""
    if definitions is None:
        definitions = schema.get("$defs", {})

    lines = toml_str.split("\n")
    commented_lines = []
    current_section = None
    current_schema = schema

    for line in lines:
        line = line.rstrip()

        # Detect section headers
        if line.startswith("[") and line.endswith("]"):
            current_section = line[1:-1]
            commented_lines.append("")
            commented_lines.append(line)

            # Navigate to the schema for this section
            current_schema = schema
            section_parts = current_section.split(".")
            for part in section_parts:
                properties = current_schema.get("properties", {})
                if part in properties:
                    field_schema = properties[part]
                    # Resolve references
                    if "$ref" in field_schema:
                        ref_name = field_schema["$ref"].split("/")[-1]
                        if ref_name in definitions:
                            current_schema = definitions[ref_name]
                    elif field_schema.get("type") == "object":
                        current_schema = field_schema
            continue

        # Skip empty lines
        if not line.strip():
            commented_lines.append(line)
            continue

        # Extract field name from key = value line
        if "=" in line:
            key = line.split("=")[0].strip()
            field_name = key.strip('"')

            # Find the field schema
            properties = current_schema.get("properties", {})
            if field_name in properties:
                field_schema = properties[field_name]

                # Resolve references for comment extraction
                if "$ref" in field_schema:
                    ref_name = field_schema["$ref"].split("/")[-1]
                    if ref_name in definitions:
                        field_schema = definitions[ref_name]

                # Add comment if description exists
                comment = _get_field_comment(field_schema)
                if comment:
                    commented_lines.append(_wrap_comment(comment))

        commented_lines.append(line)

    return "\n".join(commented_lines)


def generate_sample_config(model_class: Type[BaseModel]) -> str:
    """
    Generate a sample TOML configuration file from a Pydantic model.

    Args:
        model_class: The Pydantic model class to generate configuration for

    Returns:
        A string containing the TOML configuration with comments
    """
    # Get JSON schema from the model
    schema = model_class.model_json_schema()

    # Generate configuration dictionary from schema
    config_dict = _generate_config_from_schema(schema)

    # Generate basic TOML
    toml_str = toml.dumps(config_dict)

    # Add comments based on schema
    commented_toml = _add_field_comments_from_schema(toml_str, schema)

    return commented_toml


def generate_sample_config_file(
    model_class: Type[BaseModel], output_path: str, header_comment: Optional[str] = None
) -> None:
    """
    Generate a sample TOML configuration file and write it to disk.

    Args:
        model_class: The Pydantic model class to generate configuration for
        output_path: Path where the configuration file will be written
        header_comment: Optional header comment to add at the top of the file
    """
    config_content = generate_sample_config(model_class)

    with open(output_path, "w") as f:
        if header_comment:
            f.write(_wrap_comment(header_comment))
            f.write("\n\n")
        f.write(config_content)



from __future__ import annotations

import asyncio
import base64
import enum
import logging
import secrets
import time
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Sequence
from typing import (
    Any,
    Callable,
    Coroutine,
    Generic,
    Optional,
    Protocol,
    Type,
    TypedDict,
    TypeVar,
    Union,
    cast,
    override,
)

import attrs
from aiomonitor.task import preserve_termination_log
from aiotools.taskgroup import PersistentTaskGroup
from aiotools.taskgroup.types import AsyncExceptionHandler

from ai.backend.common.message_queue.queue import AbstractMessageQueue
from ai.backend.common.message_queue.types import (
    BroadcastMessage,
    MessageId,
    MessagePayload,
    MQMessage,
)
from ai.backend.logging import BraceStyleAdapter

from .. import msgpack
from ..types import (
    AgentId,
)
from .reporter import AbstractEventReporter, CompleteEventReportArgs, PrepareEventReportArgs
from .types import AbstractAnycastEvent, AbstractBroadcastEvent, AbstractEvent

__all__ = (
    "EventCallback",
    "EventDispatcher",
    "EventHandler",
    "EventProducer",
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class _EventHandlerType(enum.StrEnum):
    CONSUMER = "consumer"
    SUBSCRIBER = "subscriber"


TEvent = TypeVar("TEvent", bound="AbstractEvent")
TSubscirbedEvent = TypeVar("TSubscirbedEvent", bound=AbstractBroadcastEvent)
TConsumedEvent = TypeVar("TConsumedEvent", bound=AbstractAnycastEvent)
TEventCov = TypeVar("TEventCov", bound="AbstractEvent")
TContext = TypeVar("TContext")

EventCallback = Union[
    Callable[[TContext, AgentId, TEvent], Coroutine[Any, Any, None]],
    Callable[[TContext, AgentId, TEvent], None],
]


class EventHandlerType(enum.Enum):
    CONSUMER = "CONSUMER"
    SUBSCRIBER = "SUBSCRIBER"


@attrs.define(auto_attribs=True, slots=True, frozen=True, eq=False, order=False)
class EventHandler(Generic[TContext, TEvent]):
    event_cls: Type[TEvent]
    name: str
    context: TContext
    callback: EventCallback[TContext, TEvent]
    handler_type: _EventHandlerType
    coalescing_opts: Optional[CoalescingOptions]
    coalescing_state: CoalescingState
    args_matcher: Callable[[tuple], bool] | None
    event_start_reporters: tuple[AbstractEventReporter, ...] = attrs.field(factory=tuple)
    event_complete_reporters: tuple[AbstractEventReporter, ...] = attrs.field(factory=tuple)


class CoalescingOptions(TypedDict):
    max_wait: float
    max_batch_size: int


@attrs.define(auto_attribs=True, slots=True)
class CoalescingState:
    batch_size: int = 0
    last_added: float = 0.0
    last_handle: asyncio.TimerHandle | None = None
    fut_sync: asyncio.Future | None = None

    def proceed(self):
        if self.fut_sync is not None and not self.fut_sync.done():
            self.fut_sync.set_result(None)

    async def rate_control(self, opts: CoalescingOptions | None) -> bool:
        if opts is None:
            return True
        loop = asyncio.get_running_loop()
        if self.fut_sync is None:
            self.fut_sync = loop.create_future()
        assert self.fut_sync is not None
        self.last_added = loop.time()
        self.batch_size += 1
        if self.batch_size >= opts["max_batch_size"]:
            assert self.last_handle is not None
            self.last_handle.cancel()
            self.fut_sync.cancel()
            self.last_handle = None
            self.last_added = 0.0
            self.batch_size = 0
            return True
        # Schedule.
        self.last_handle = loop.call_later(
            opts["max_wait"],
            self.proceed,
        )
        if self.last_added > 0 and loop.time() - self.last_added < opts["max_wait"]:
            # Cancel the previously pending task.
            self.last_handle.cancel()
            self.fut_sync.cancel()
            # Reschedule.
            self.fut_sync = loop.create_future()
            self.last_handle = loop.call_later(
                opts["max_wait"],
                self.proceed,
            )
        try:
            await self.fut_sync
        except asyncio.CancelledError:
            if self.last_handle is not None and not self.last_handle.cancelled():
                self.last_handle.cancel()
            return False
        else:
            self.fut_sync = None
            self.last_handle = None
            self.last_added = 0.0
            self.batch_size = 0
            return True


class EventObserver(Protocol):
    def observe_event_success(self, *, event_type: str, duration: float) -> None: ...

    def observe_event_failure(
        self, *, event_type: str, duration: float, exception: BaseException
    ) -> None: ...


class NopEventObserver:
    def observe_event_success(self, *, event_type: str, duration: float) -> None:
        pass

    def observe_event_failure(
        self, *, event_type: str, duration: float, exception: BaseException
    ) -> None:
        pass


class _ConsumerPostCallback:
    def __init__(
        self,
        msg_id: MessageId,
        msg_queue: AbstractMessageQueue,
        remaining_handler_cnt: int,
    ) -> None:
        self._msg_id = msg_id
        self._msg_queue = msg_queue
        self._remaining_handler_cnt = remaining_handler_cnt
        self._lock = asyncio.Lock()

    async def done(self) -> None:
        # To ensure that all consumer handlers are called.
        # Basically there should be only one consumer handler for one event.
        async with self._lock:
            self._remaining_handler_cnt -= 1
            if self._remaining_handler_cnt > 0:
                return
        # All consumer handlers are called.
        await self._msg_queue.done(self._msg_id)


class PostCallback(Protocol):
    async def done(self) -> None:
        pass


class EventDispatcherGroup(ABC):
    @abstractmethod
    def with_reporters(
        self,
        start_reporters: Sequence[AbstractEventReporter] = tuple(),
        complete_reporters: Sequence[AbstractEventReporter] = tuple(),
    ) -> EventDispatcherGroup:
        raise NotImplementedError

    @abstractmethod
    def consume(
        self,
        event_cls: Type[TConsumedEvent],
        context: TContext,
        callback: EventCallback[TContext, TConsumedEvent],
        coalescing_opts: Optional[CoalescingOptions] = None,
        *,
        name: Optional[str] = None,
        args_matcher: Optional[Callable[[tuple], bool]] = None,
    ) -> EventHandler[TContext, TConsumedEvent]:
        raise NotImplementedError

    @abstractmethod
    def subscribe(
        self,
        event_cls: Type[TSubscirbedEvent],
        context: TContext,
        callback: EventCallback[TContext, TSubscirbedEvent],
        coalescing_opts: Optional[CoalescingOptions] = None,
        *,
        name: Optional[str] = None,
        override_event_name: Optional[str] = None,
        args_matcher: Optional[Callable[[tuple], bool]] = None,
    ) -> EventHandler[TContext, TSubscirbedEvent]:
        raise NotImplementedError


class _EventDispatcherWrapper(EventDispatcherGroup):
    _event_dispatcher: EventDispatcher

    _start_reporters: list[AbstractEventReporter]
    _complete_reporters: list[AbstractEventReporter]

    def __init__(
        self,
        event_dispatcher: EventDispatcher,
        start_reporters: Sequence[AbstractEventReporter] = tuple(),
        complete_reporters: Sequence[AbstractEventReporter] = tuple(),
    ) -> None:
        self._event_dispatcher = event_dispatcher
        self._start_reporters = list(start_reporters)
        self._complete_reporters = list(complete_reporters)

    @override
    def with_reporters(
        self,
        start_reporters: Sequence[AbstractEventReporter] = tuple(),
        complete_reporters: Sequence[AbstractEventReporter] = tuple(),
    ) -> _EventDispatcherWrapper:
        return _EventDispatcherWrapper(
            event_dispatcher=self._event_dispatcher,
            start_reporters=self._start_reporters + list(start_reporters),
            complete_reporters=self._complete_reporters + list(complete_reporters),
        )

    @override
    def consume(
        self,
        event_cls: Type[TConsumedEvent],
        context: TContext,
        callback: EventCallback[TContext, TConsumedEvent],
        coalescing_opts: Optional[CoalescingOptions] = None,
        *,
        name: Optional[str] = None,
        args_matcher: Optional[Callable[[tuple], bool]] = None,
    ) -> EventHandler[TContext, TConsumedEvent]:
        return self._event_dispatcher.consume(
            event_cls,
            context,
            callback,
            coalescing_opts=coalescing_opts,
            name=name,
            args_matcher=args_matcher,
            start_reporters=tuple(self._start_reporters),
            complete_reporters=tuple(self._complete_reporters),
        )

    @override
    def subscribe(
        self,
        event_cls: Type[TSubscirbedEvent],
        context: TContext,
        callback: EventCallback[TContext, TSubscirbedEvent],
        coalescing_opts: Optional[CoalescingOptions] = None,
        *,
        name: Optional[str] = None,
        override_event_name: Optional[str] = None,
        args_matcher: Optional[Callable[[tuple], bool]] = None,
    ) -> EventHandler[TContext, TSubscirbedEvent]:
        return self._event_dispatcher.subscribe(
            event_cls,
            context,
            callback,
            coalescing_opts=coalescing_opts,
            name=name,
            override_event_name=override_event_name,
            args_matcher=args_matcher,
            start_reporters=tuple(self._start_reporters),
            complete_reporters=tuple(self._complete_reporters),
        )


class EventDispatcher(EventDispatcherGroup):
    """
    We have two types of event handlers: consumer and subscriber.

    Consumers use the distribution pattern. Only one consumer among many manager worker processes
    receives the event.

    Consumer example: database updates upon specific events.

    Subscribers use the broadcast pattern. All subscribers in many manager worker processes
    receive the same event.

    Subscriber example: enqueuing events to the queues for event streaming API handlers
    """

    _consumers: defaultdict[
        str, set[EventHandler[Any, AbstractEvent]]
    ]  # TODO: set only one consumer handler for one event
    _subscribers: defaultdict[str, set[EventHandler[Any, AbstractEvent]]]
    _msg_queue: AbstractMessageQueue

    _consumer_loop_task: Optional[asyncio.Task]
    _subscriber_loop_task: Optional[asyncio.Task]
    _consumer_taskgroup: PersistentTaskGroup
    _subscriber_taskgroup: PersistentTaskGroup

    _log_events: bool
    _metric_observer: EventObserver

    def __init__(
        self,
        message_queue: AbstractMessageQueue,
        log_events: bool = False,
        *,
        consumer_exception_handler: AsyncExceptionHandler | None = None,
        subscriber_exception_handler: AsyncExceptionHandler | None = None,
        event_observer: EventObserver = NopEventObserver(),
    ) -> None:
        self._log_events = log_events
        self._closed = False
        self._consumers = defaultdict(set)
        self._subscribers = defaultdict(set)
        self._msg_queue = message_queue
        self._metric_observer = event_observer
        self._consumer_taskgroup = PersistentTaskGroup(
            name="consumer_taskgroup",
            exception_handler=consumer_exception_handler,
        )
        self._subscriber_taskgroup = PersistentTaskGroup(
            name="subscriber_taskgroup",
            exception_handler=subscriber_exception_handler,
        )
        self._consumer_loop_task = None
        self._subscriber_loop_task = None

    async def start(self) -> None:
        if self._closed:
            return
        self._consumer_loop_task = asyncio.create_task(self._consume_loop())
        self._subscriber_loop_task = asyncio.create_task(self._subscribe_loop())

    async def close(self) -> None:
        self._closed = True
        try:
            cancelled_tasks = []
            await self._consumer_taskgroup.shutdown()
            await self._subscriber_taskgroup.shutdown()

            def cancel_task(task: Optional[asyncio.Task]) -> None:
                if task is not None and not task.done():
                    task.cancel()
                    cancelled_tasks.append(task)

            cancel_task(self._consumer_loop_task)
            cancel_task(self._subscriber_loop_task)
            await asyncio.gather(*cancelled_tasks, return_exceptions=True)
        except Exception:
            log.exception("unexpected error while closing event dispatcher")

    @override
    def with_reporters(
        self,
        start_reporters: Sequence[AbstractEventReporter] = tuple(),
        complete_reporters: Sequence[AbstractEventReporter] = tuple(),
    ) -> EventDispatcherGroup:
        return _EventDispatcherWrapper(
            event_dispatcher=self,
            start_reporters=list(start_reporters),
            complete_reporters=list(complete_reporters),
        )

    @override
    def consume(
        self,
        event_cls: Type[TConsumedEvent],
        context: TContext,
        callback: EventCallback[TContext, TConsumedEvent],
        coalescing_opts: Optional[CoalescingOptions] = None,
        *,
        name: Optional[str] = None,
        args_matcher: Optional[Callable[[tuple], bool]] = None,
        start_reporters: Sequence[AbstractEventReporter] = tuple(),
        complete_reporters: Sequence[AbstractEventReporter] = tuple(),
    ) -> EventHandler[TContext, TConsumedEvent]:
        """
        Register a callback as a consumer. When multiple callback registers as a consumer
        on a single event, only one callable among those will be called.

        args_matcher:
          Optional. A callable which accepts event argument and supplies a bool as a return value.
          When specified, EventDispatcher will only execute callback when this lambda returns True.
        """

        if name is None:
            name = f"evh-{secrets.token_urlsafe(16)}"
        handler = EventHandler(
            event_cls,
            name,
            context,
            callback,
            _EventHandlerType.CONSUMER,
            coalescing_opts,
            CoalescingState(),
            args_matcher,
            event_start_reporters=tuple(start_reporters),
            event_complete_reporters=tuple(complete_reporters),
        )
        self._consumers[event_cls.event_name()].add(cast(EventHandler[Any, AbstractEvent], handler))
        return handler

    def unconsume(
        self,
        handler: EventHandler[TContext, TEvent],
    ) -> None:
        self._consumers[handler.event_cls.event_name()].discard(
            cast(EventHandler[Any, AbstractEvent], handler)
        )

    @override
    def subscribe(
        self,
        event_cls: Type[TSubscirbedEvent],
        context: TContext,
        callback: EventCallback[TContext, TSubscirbedEvent],
        coalescing_opts: Optional[CoalescingOptions] = None,
        *,
        name: Optional[str] = None,
        override_event_name: Optional[str] = None,
        args_matcher: Optional[Callable[[tuple], bool]] = None,
        start_reporters: Sequence[AbstractEventReporter] = tuple(),
        complete_reporters: Sequence[AbstractEventReporter] = tuple(),
    ) -> EventHandler[TContext, TSubscirbedEvent]:
        """
        Subscribes to given event. All handlers will be called when certain event pops up.

        args_matcher:
          Optional. A callable which accepts event argument and supplies a bool as a return value.
          When specified, EventDispatcher will only execute callback when this lambda returns True.
        """

        if name is None:
            name = f"evh-{secrets.token_urlsafe(16)}"
        handler = EventHandler(
            event_cls,
            name,
            context,
            callback,
            _EventHandlerType.SUBSCRIBER,
            coalescing_opts,
            CoalescingState(),
            args_matcher,
            event_start_reporters=tuple(start_reporters),
            event_complete_reporters=tuple(complete_reporters),
        )
        override_event_name = override_event_name or event_cls.event_name()
        self._subscribers[override_event_name].add(cast(EventHandler[Any, AbstractEvent], handler))
        return handler

    def unsubscribe(
        self,
        handler: EventHandler[TContext, TEvent],
        *,
        override_event_name: Optional[str] = None,
    ) -> None:
        override_event_name = override_event_name or handler.event_cls.event_name()
        self._subscribers[override_event_name].discard(
            cast(EventHandler[Any, AbstractEvent], handler)
        )

    async def _handle(
        self,
        evh: EventHandler,
        source: AgentId,
        args: tuple,
        post_callbacks: Sequence[PostCallback] = tuple(),
    ) -> None:
        if evh.args_matcher and not evh.args_matcher(args):
            return
        coalescing_opts = evh.coalescing_opts
        coalescing_state = evh.coalescing_state
        cb = evh.callback
        evh_type = evh.handler_type
        event_cls = evh.event_cls
        if self._closed:
            return
        event_type = event_cls.event_name()
        event = event_cls.deserialize(args)
        start = time.perf_counter()
        for start_reporter in evh.event_start_reporters:
            await start_reporter.prepare_event_report(event, PrepareEventReportArgs())
        try:
            if await coalescing_state.rate_control(coalescing_opts):
                if self._closed:
                    return
                if self._log_events:
                    log.debug("DISPATCH_{}(evh:{})", evh_type.name, evh.name)
                if asyncio.iscoroutinefunction(cb):
                    # mypy cannot catch the meaning of asyncio.iscoroutinefunction().
                    await cb(evh.context, source, event)  # type: ignore
                else:
                    cb(evh.context, source, event)  # type: ignore
                for post_callback in post_callbacks:
                    await post_callback.done()
                self._metric_observer.observe_event_success(
                    event_type=event_type,
                    duration=time.perf_counter() - start,
                )
        except Exception as e:
            self._metric_observer.observe_event_failure(
                event_type=event_type,
                duration=time.perf_counter() - start,
                exception=e,
            )
            log.exception(f"EventDispatcher.{evh_type}(): unexpected-error, {repr(e)}")
            raise
        except BaseException as e:
            self._metric_observer.observe_event_failure(
                event_type=event_type,
                duration=time.perf_counter() - start,
                exception=e,
            )
            raise
        duration = time.perf_counter() - start
        for complete in evh.event_complete_reporters:
            await complete.complete_event_report(event, CompleteEventReportArgs(duration))

    async def dispatch_consumers(
        self,
        event_name: str,
        source: AgentId,
        args: tuple,
        post_callbacks: Sequence[PostCallback] = tuple(),
    ) -> None:
        if self._log_events:
            log.debug("DISPATCH_CONSUMERS(ev:{}, ag:{})", event_name, source)
        consumers_handlers = self._consumers[event_name].copy()
        if not consumers_handlers:
            # If there are no consumer handlers, we can just call post callbacks and return.
            for post_callback in post_callbacks:
                await post_callback.done()
            return
        for consumer in consumers_handlers:
            self._consumer_taskgroup.create_task(
                self._handle(consumer, source, args, post_callbacks),
            )
            await asyncio.sleep(0)

    async def dispatch_subscribers(
        self,
        event_name: str,
        source: AgentId,
        args: tuple,
    ) -> None:
        if self._log_events:
            log.debug("DISPATCH_SUBSCRIBERS(ev:{}, ag:{})", event_name, source)
        for subscriber in self._subscribers[event_name].copy():
            self._subscriber_taskgroup.create_task(
                self._handle(subscriber, source, args),
            )
            await asyncio.sleep(0)

    @preserve_termination_log
    async def _consume_loop(self) -> None:
        async for msg in self._msg_queue.consume_queue():  # type: ignore
            if self._closed:
                return
            try:
                mq_msg = cast(MQMessage, msg)
                msg_payload = MessagePayload.from_anycast(mq_msg.payload)
                post_callback = _ConsumerPostCallback(
                    mq_msg.msg_id,
                    self._msg_queue,
                    len(self._consumers[msg_payload.name]),
                )
                await self.dispatch_consumers(
                    msg_payload.name,
                    AgentId(msg_payload.source),
                    msg_payload.args,
                    [post_callback],
                )
            except Exception as e:
                log.exception(
                    "EventDispatcher._consume_loop: unexpected-error, {}",
                    repr(e),
                )
                # Do not raise the exception to avoid stopping the loop.
                # The exception will be handled by the task group.

    @preserve_termination_log
    async def _subscribe_loop(self) -> None:
        async for msg in self._msg_queue.subscribe_queue():  # type: ignore
            if self._closed:
                return
            try:
                msg = cast(BroadcastMessage, msg)
                msg_payload = MessagePayload.from_broadcast(msg.payload)
                await self.dispatch_subscribers(
                    msg_payload.name,
                    AgentId(msg_payload.source),
                    msg_payload.args,
                )
            except Exception as e:
                log.exception(
                    "EventDispatcher._subscribe_loop: unexpected-error, {}",
                    repr(e),
                )
                # Do not raise the exception to avoid stopping the loop.
                # The exception will be handled by the task group.


class EventProducer:
    _closed: bool
    _msg_queue: AbstractMessageQueue
    _source: AgentId
    _source_bytes: bytes
    _log_events: bool

    def __init__(
        self,
        msg_queue: AbstractMessageQueue,
        *,
        source: AgentId,
        log_events: bool = False,
    ) -> None:
        self._closed = False
        self._msg_queue = msg_queue
        self._source = source
        self._source_bytes = source.encode()
        self._log_events = log_events

    async def close(self) -> None:
        self._closed = True
        await self._msg_queue.close()

    async def anycast_event(
        self,
        event: AbstractAnycastEvent,
        source_override: Optional[AgentId] = None,
    ) -> None:
        if self._closed:
            return
        source_bytes = self._source_bytes
        if source_override is not None:
            source_bytes = source_override.encode()

        raw_event = {
            b"name": event.event_name().encode(),
            b"source": source_bytes,
            b"args": msgpack.packb(event.serialize()),
        }
        await self._msg_queue.send(raw_event)

    async def broadcast_event(
        self,
        event: AbstractBroadcastEvent,
        source_override: Optional[AgentId] = None,
    ) -> None:
        if self._closed:
            return
        source = self._source
        if source_override is not None:
            source = source_override
        args = base64.b64encode(msgpack.packb(event.serialize())).decode("ascii")
        raw_event = {
            "name": event.event_name(),
            "source": source,
            "args": args,
        }
        await self._msg_queue.broadcast(raw_event)

    async def broadcast_event_with_cache(
        self,
        cache_id: str,
        event: AbstractBroadcastEvent,
    ) -> None:
        """
        Broadcast a message to all subscribers with cache.
        The message will be delivered to all subscribers.
        """
        args = base64.b64encode(msgpack.packb(event.serialize())).decode("ascii")
        raw_event = {
            "name": event.event_name(),
            "source": str(self._source),
            "args": args,
        }
        await self._msg_queue.broadcast_with_cache(
            cache_id,
            raw_event,
        )

    async def anycast_and_broadcast_event(
        self,
        anycast_event: AbstractAnycastEvent,
        broadcast_event: AbstractBroadcastEvent,
    ) -> None:
        """
        Send both anycast and broadcast events.

        Note:
        Recommend to use `anycast_event` and `broadcast_event` separately.
        Do not use this method as long as possible.
        """
        await self.anycast_event(anycast_event)
        await self.broadcast_event(broadcast_event)
